#include "gis.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

/* ISSUES

	- make sure that pathnames are platform independent
	- make sure that Gmakefile will work on other platforms

*/

int
main (int argc, char *argv[])
{
	struct GModule *module;
	struct
	{
		struct Option *file;
	}
	parm;
	struct
	{
		struct Flag *quiet;
		struct Flag *gui;
	}
	flag;
	
	FILE *kb; /* knowledge base */
	
	xmlDocPtr doc;
	xmlNodePtr cur;
	xmlNodePtr root;
	xmlNodePtr subroot;
	int retval = 0;
	char xmlfile [255];

	
	/* setup some basic GIS stuff */
	G_gisinit (argv[0]);
	module = G_define_module ();
	module->description = "Create a new Dempster-Shafer knowledge base.";
	
	/* do not pause after a warning message was displayes */
	G_sleep_on_error (0);

	/* Parameters */
	parm.file = G_define_option ();
	parm.file->key = "file";
	parm.file->type = TYPE_STRING;
	parm.file->required = YES;
	parm.file->description = "filename for the knowledge base.";

	/* parse command line */
	if (G_parser (argc, argv))
	{
		exit (-1);
	}

	/* check if given parameters are valid */
	if (G_legal_filename (parm.file->answer) == -1) {
		G_fatal_error ("Please provide a legal filename!\n");
	}
	
	/* existing files will be overwritten! */
	kb = G_fopen_new ("DST",parm.file->answer);
	
	if ( kb == NULL ) {
		G_fatal_error ("Cannot open knowledge base file for writing!\n");
	}
	
	/* get a fully qualified, absolute path and filename for */
	/* writing the XML database file. */
	/* What to do on a DOS system? Let's hope that libxml handles it!*/
	sprintf (xmlfile,"%s/%s/DST/%s",G_location_path(),G_mapset(),parm.file->answer);
	
	fclose(kb);
	
	/* create basic XML structure */
	root = xmlNewNode(NULL, (xmlChar *) "dst-kb");
	xmlSetProp(root,(xmlChar*) "N_SINGLETONS", (xmlChar*) "0");
	/* the NULL hypotheses (empty set) is implicit */
	xmlSetProp(root,(xmlChar*) "N_HYPOTHESES", (xmlChar*) "1");
	/* initial number of groups in the knowledge base */
	xmlSetProp(root,(xmlChar*) "N_GROUPS", (xmlChar*) "0");
	
	xmlNewTextChild (root, NULL, "hyp", "NULL");
	
	doc = xmlNewDoc("1.0");
	xmlDocSetRootElement(doc, root);
	cur = xmlDocGetRootElement(doc);
	cur = cur->xmlChildrenNode;
	xmlSetProp(cur,(xmlChar*) "TYPE", (xmlChar*) "NULL");
	
	if (xmlSaveFormatFile (xmlfile, doc, 1) == -1) {
		G_fatal_error ("Error creating XML file!\n");
	}	
	
	return 0;
}
